from unittest import mock

from docker.errors import NotFound

from teststack import cli


def test_env_with_containers_inside(runner, attrs):
    client = mock.MagicMock()
    client.containers.get.return_value.attrs = attrs
    with mock.patch('docker.from_env', return_value=client):
        result = runner.invoke(cli, ['env', '--inside'])
    assert result.exit_code == 0
    assert 'AWS_ACCESS_KEY_ID' in result.output
    assert 'POSTGRES_MAIN_DBNAME=bebop' in result.output
    assert 'POSTGRES_MAIN_HOST=fakeaddress' in result.output
    assert 'POSTGRES_MAIN_PORT=5432' in result.output


def test_env_with_containers_outside(runner, attrs):
    client = mock.MagicMock()
    client.containers.get.return_value.attrs = attrs
    with mock.patch('docker.from_env', return_value=client):
        result = runner.invoke(cli, ['env'])
    assert result.exit_code == 0
    assert 'AWS_ACCESS_KEY_ID' in result.output
    assert 'POSTGRES_MAIN_DBNAME=bebop' in result.output
    assert 'POSTGRES_MAIN_HOST=localhost' in result.output
    assert 'POSTGRES_MAIN_PORT=12345' in result.output


def test_env_without_containers(runner, attrs):
    client = mock.MagicMock()
    client.containers.get.return_value.attrs = attrs
    client.containers.get.side_effect = NotFound('container not found')
    with mock.patch('docker.from_env', return_value=client):
        result = runner.invoke(cli, ['env'])
    assert result.exit_code == 0
    assert 'AWS_ACCESS_KEY_ID' in result.output
    assert 'POSTGRES_MAIN_DBNAME=bebop' not in result.output
    assert 'POSTGRES_MAIN_HOST=localhost' not in result.output
    assert 'POSTGRES_MAIN_PORT=12345' not in result.output


def test_env_without_containers_quiet(runner):
    client = mock.MagicMock()
    client.containers.get.side_effect = NotFound('container not found')
    with mock.patch('docker.from_env', return_value=client):
        result = runner.invoke(cli, ['env', '--quiet'])
    assert result.exit_code == 0
    assert 'AWS_ACCESS_KEY_ID' not in result.output



import pathlib

import click.testing
import docker as docker_py

import pytest

from teststack import import_commands
from teststack.git import get_tag


@pytest.fixture
def runner():
    import_commands()
    return click.testing.CliRunner()


@pytest.fixture()
def tag():
    return get_tag()


@pytest.fixture()
def attrs():
    return {
        'NetworkSettings': {
            'IPAddress': 'fakeaddress',
            'Ports': {
                '5432/tcp': [
                    {'HostPort': '12345'},
                ],
                '5672/tcp': [
                    {'HostPort': '12537'},
                ],
                '12345/tcp': [],
            },
        },
    }


@pytest.fixture()
def build_output():
    chunks = []
    with open('tests/files/build.output', 'rb') as fh_:
        for line in fh_:
            line = line.replace(b'\\r\\n', b'\r\n')
            line = line.replace(b'\\\\', b'\\')
            chunks.append(line)
    return chunks


@pytest.fixture()
def main_dir():
    return pathlib.Path(__file__).parent.parent


@pytest.fixture()
def testapp_dir(main_dir):
    return main_dir / 'tests' / 'testapp'


@pytest.fixture()
def docker():
    return docker_py.from_env()


@pytest.fixture(autouse=True)
def test_no_leftover_docker_containers(docker):
    yield
    assert not any(
        container.name.startswith('testapp') for container in docker.containers.list()
    ), '`testapp` containers were left behind, please clean them up in this test'
    assert not any(
        container.name.startswith('teststack') for container in docker.containers.list()
    ), '`teststack` containers were left behind, please clean them up in this test'



"""
The containers related commands handle the majority of the work that teststack does.

The work is abstracted into the different container drivers.  You can specify
which container driver to use in the ``teststack.toml`` file.

.. code-block:: toml

    [client]
    name = "docker"

All of the commands can be combined together and they will be executed in order.
So if for example you want to rebuild the image without any cache and then run
the tests you could do the following.

.. code-block::

    teststack build --rebuild run
"""
import os.path
import sys

import click
import jinja2

from teststack import cli


@cli.command()
@click.option('--no-tests', '-n', is_flag=True, help='Don\'t start the tests container')
@click.pass_context
def start(ctx, no_tests):
    """
    Start services and tests containers.

    If the flag ``--no-tests`` is passed, then the tests container is not
    started, and only the services are started. This is useful if running tests
    outside of a docker container.

    --no-tests, -n

        do not build an image or start a tests container

    .. code-block:: bash

        teststack start --no-tests
    """
    client = ctx.obj['client']
    for service, data in ctx.obj['services'].items():
        name = f'{ctx.obj["project_name"]}_{service}'
        container = client.container_get(name)
        if container is None:
            client.run(
                image=data['image'],
                ports=data.get('ports', {}),
                name=name,
                command=data.get('command', None),
                environment=data.get('environment', {}),
                mount_cwd=False,
            )
        else:
            client.start(name=name)

    if no_tests is True:
        return

    env = ctx.invoke(cli.get_command(ctx, 'env'), inside=True, no_export=True, quiet=True)
    env = dict(line.split('=') for line in env)
    image = client.image_get(ctx.obj['tag'])
    if image is None:
        image = client.image_get(ctx.invoke(build))

    name = f'{ctx.obj["project_name"]}_tests'

    current_image_id = client.container_get_current_image(name)
    if current_image_id != image:
        client.end_container(name)
        current_image_id = None
    else:
        container = client.container_get(name)

    if current_image_id is None:
        command = ctx.obj['tests'].get('command', True)

        container = client.run(
            image=image,
            stream=True,
            name=name,
            environment=env,
            command=command,
            mount_cwd=True,
        )

    return container


@cli.command()
@click.pass_context
def stop(ctx):
    """
    Stop all containers

    .. code-block:: bash

        teststack stop
    """
    client = ctx.obj['client']
    project_name = ctx.obj["project_name"]
    for service, _ in ctx.obj['services'].items():
        name = f'{project_name}_{service}'
        container = client.container_get(name)
        if container is None:
            continue
        client.end_container(container)
    container = client.container_get(f'{project_name}_tests')
    if container is None:
        return
    client.end_container(container)


@cli.command()
@click.pass_context
def restart(ctx):  # pragma: no cover
    """
    Stop then Start all containers

    .. code-block:: bash

        teststack restart
    """
    ctx.invoke(stop)
    ctx.invoke(start)


@cli.command()
@click.option(
    '--template-file',
    '-t',
    type=click.File(),
    default='Dockerfile.j2',
    help='template to render with jinja',
)
@click.option(
    '--dockerfile', '--file', '-f', type=click.Path(), default='Dockerfile', help='container build file to write to'
)
@click.pass_context
def render(ctx, template_file, dockerfile):
    """
    Render the template_file to the dockerfile.

    --template-file, -t

        template file to render, default: Dockerfile.j2

    --dockerfile, --file, -f

        file to write the rendered dockerfile to. default: Dockerfile

    .. code-block:: bash

        teststack render
        teststack render --template-file Containerfile.j2 --file Containerfile
    """
    env = jinja2.Environment(
        extensions=[
            'jinja2.ext.i18n',
            'jinja2.ext.do',
            'jinja2.ext.loopcontrols',
        ],
        keep_trailing_newline=True,
        undefined=jinja2.Undefined,
        loader=jinja2.FileSystemLoader(os.getcwd()),
    )

    template_string = template_file.read()

    if 'commit' in ctx.obj:
        template_string = '\n'.join(
            [
                template_string,
                f'RUN echo "app-git-hash: {ctx.obj["commit"]} >> /etc/docker-metadata"',
                f'ENV APP_GIT_HASH={ctx.obj["commit"]}\n',
            ]
        )

    template = env.from_string(
        '\n'.join(
            [
                template_string,
            ]
        ),
    )
    template.stream(
        GIT_BRANCH=ctx.obj.get('branch', 'dev'),
        GIT_COMMIT_HASH=ctx.obj.get('commit', None),
        **os.environ,
    ).dump(dockerfile)


@cli.command()
@click.option('--rebuild', '-r', is_flag=True, help='ignore cache and rebuild the container fully')
@click.option('--tag', '-t', default=None, help='Tag to label the build')
@click.option(
    '--dockerfile', '--file', '-f', type=click.Path(), default='Dockerfile', help='container build file to write too'
)
@click.option('--template-file', type=click.Path(), default='Dockerfile.j2', help='template to render with jinja')
@click.pass_context
def build(ctx, rebuild, tag, dockerfile, template_file):
    """
    Build the docker image using the dockerfile.

    If the dockerfile does not exist, then it will be rendered.

    --template-file

        Template to render to the Dockerfile

    --dockerfile, --file, -f

        dockerfile to build into an image. Default: Dockerfile

    --rebuild, -r

        Ignore the cache and rebuild from scratch

    --tag, -t

        Tag to use for the image.  Default: <dirname>:<latest git hash/"latest">

    .. code-block:: bash

        teststack build
        teststack build --tag blah:old
    """
    try:
        tempstat = os.stat(template_file)
    except FileNotFoundError:
        tempstat = None

    try:
        dockerstat = os.stat(dockerfile)
    except FileNotFoundError:
        dockerstat = None

    if tempstat is not None and (dockerstat is None or dockerstat.st_mtime < tempstat.st_mtime):
        with open(template_file, 'r') as th_:
            ctx.invoke(render, dockerfile=dockerfile, template_file=th_)

    client = ctx.obj['client']

    if tag is None:
        tag = ctx.obj['tag']

    click.echo(f'Build Image: {tag}')
    client.build(dockerfile, tag, rebuild)
    image = client.image_get(tag)
    if image is None:
        click.echo(click.style('Failed to build image!', fg='red'))
        sys.exit(11)

    return tag


@cli.command()
@click.pass_context
def exec(ctx):  # pragma: no cover
    """
    Exec into the current tests container.

    .. code-block:: bash

        teststack exec
    """
    container = ctx.invoke(start)
    ctx.obj['client'].exec(container)


@cli.command()
@click.pass_context
def tag(ctx):
    click.echo(ctx.obj['tag'])


@cli.command()
@click.option('--step', '-s', help='Which step to run')
@click.argument('posargs', nargs=-1, type=click.UNPROCESSED)
@click.pass_context
def run(ctx, step, posargs):
    """
    Run the specified test steps from the teststack.toml.

    --step, -s

        specify a single step to run.

    posargs

        All other leftover unprocessed arguments are passed as {posargs} to be
        rendered when the commands.

    .. code-block:: bash

        teststack run
        teststack run --step tests -- -k test_add_user tests/unit/test_users.py
    """
    container = ctx.invoke(start)
    client = ctx.obj['client']

    steps = ctx.obj['tests'].get('steps', {})
    if step:
        commands = [steps.get(step, '{posargs}')]
    else:
        commands = steps.values()
    for command in commands:
        if isinstance(command, list):
            for cmd in command:
                client.run_command(container, cmd.format(posargs=' '.join(posargs)))
        else:
            client.run_command(container, command.format(posargs=' '.join(posargs)))



import typing

import flask
import flask_caching

from .database import db
from .handlers.v1 import blueprint as blueprint_v1


def create_app(config: typing.Dict = None) -> flask.Flask:
    app = flask.Flask(__name__)
    app.config.from_object('app.config:Config')
    if config is not None:
        app.config.update(config)
    db.init_app(app)
    flask_caching.Cache(app)
    app.register_blueprint(blueprint_v1)
    return app



import sys

import toml

import pytest

from teststack import cli
from teststack import import_commands
from teststack.errors import IncompatibleVersionError


def test_import_commands():
    import_commands()
    assert "teststack.commands.containers" in sys.modules
    assert "teststack.commands.environment" in sys.modules


def test_min_version(runner):
    with runner.isolated_filesystem() as th_:
        with open(f'{th_}/teststack.toml', 'w') as fh_:
            toml.dump({'tests': {'min_version': 'v999.999.999'}}, fh_)
        result = runner.invoke(cli, [f'--path={th_}', 'env'])
        assert result.exit_code == 10



import os.path
import pathlib
import sys
from distutils.version import LooseVersion

import click
import toml

from . import git
from .errors import IncompatibleVersionError

try:
    from importlib.metadata import entry_points
except ImportError:
    from backports.entry_points_selectable import entry_points

try:
    from ._version import version as __version__
except ImportError:  # pragma: no cover
    try:
        from setuptools_scm import get_version

        __version__ = get_version(root='..', relative_to=__file__)
    except ImportError:
        __version__ = 'v0.0.1'


@click.group(chain=True)
@click.option(
    '--config',
    '-c',
    type=click.Path(),
    default='teststack.toml',
    help='Location of teststack config.',
)
@click.option(
    '--project-name',
    '-n',
    default=None,
    help='Prefix for docker objects.',
)
@click.option('--path', '-p', default=os.getcwd(), type=click.Path(exists=True), help='Directory to run teststack in.')
@click.pass_context
def cli(ctx, config, project_name, path):
    ctx.ensure_object(dict)
    config = pathlib.Path(config)

    # change dir before everything else is calculated
    ctx.obj['currentdir'] = os.getcwd()
    os.chdir(path)

    if config.exists():
        with config.open('r') as fh_:
            config = toml.load(fh_)
    else:
        config = {}

    min_version = LooseVersion(config.get('tests', {}).get('min_version', 'v0.0.0').lstrip('v'))
    if min_version > LooseVersion(__version__):
        click.echo(f'Current teststack version is too low, upgrade to atleast {min_version}', err=True)
        sys.exit(10)

    ctx.obj['services'] = config.get('services', {})
    ctx.obj['tests'] = config.get('tests', {})
    ctx.obj['project_name'] = os.path.basename(path.strip('/')) if project_name is None else project_name

    ctx.obj['client'] = get_client(config.get('client', {}))
    ctx.obj.update(git.get_tag(prefix=config.get('client', {}).get('prefix', '')))


def get_client(client):
    group = 'teststack.clients'
    entries = entry_points()

    if hasattr(entries, 'select'):
        entries = entries.select(group=group)
    else:
        entries = entries.get(group, [])

    client_name = client.pop('name', 'docker')
    for entry_point in entries:
        if entry_point.name == client_name:
            return entry_point.load().Client(**client)


def import_commands():
    group = 'teststack.commands'
    entries = entry_points()

    if hasattr(entries, 'select'):
        entries = entries.select(group=group)
    else:
        entries = entries.get(group, [])

    for entry_point in entries:
        entry_point.load()


def main():  # pragma: no cover
    import_commands()
    cli()



import click

from teststack import cli


@cli.command()
@click.option(
    '--no-export',
    '-n',
    is_flag=True,
    default=False,
    help='do not include the export command with environment variables',
)
@click.option('--inside', is_flag=True, default=False, help='Export variables for inside a docker container')
@click.option('--quiet', '-q', is_flag=True, help='Do not print out information')
@click.pass_context
def env(ctx, no_export, inside, quiet):
    """
    Output the environment variables for the teststack environment.

    --no-export, -n

        Do not prefix each line with export, this is good for making .env files for
        stuff like VSCode

    --inside

        export HOST and PORT values for inside of the containers. This is used
        to set the environment variables in the ``tests`` container.

    --quiet, -q

        Do not print anything out, just return the environment variables. This
        is for internal use so that the variables are not printed out when
        called inside teststack
    """
    envvars = []
    client = ctx.obj['client']
    for service, data in ctx.obj['services'].items():
        name = f'{ctx.obj["project_name"]}_{service}'
        container_data = client.get_container_data(name, inside=inside)
        if container_data is None:
            continue
        container_data.update(data.get('environment', {}).copy())
        for key, value in data.get('export', {}).items():
            envvars.append(
                f'{"" if no_export else "export "}{key}={value}'.format_map(
                    container_data,
                )
            )
    for key, value in ctx.obj['tests'].get('environment', {}).items():
        envvars.append(f'{"" if no_export else "export "}{key}={value}')
    if quiet is False:
        click.echo('\n'.join(envvars))
    return envvars



import contextlib
import sys
import termios
import tty


class read_from_stdin:
    def __enter__(self):
        if sys.stdin.isatty():  # pragma: no cover
            fd = sys.stdin.fileno()
            self.orig_fl = termios.tcgetattr(fd)
            tty.setcbreak(fd)  # use tty.setraw() instead to catch ^C also
            mode = termios.tcgetattr(fd)
            CC = 6
            mode[CC][termios.VMIN] = 0
            mode[CC][termios.VTIME] = 0
            termios.tcsetattr(fd, termios.TCSAFLUSH, mode)

            return fd

        return None

    def __exit__(self, exc_type, exc_val, traceback):
        if getattr(self, 'orig_fl', None) is not None:
            termios.tcsetattr(sys.stdin.fileno(), termios.TCSANOW, self.orig_fl)



import flask_sqlalchemy


class BaseModel(flask_sqlalchemy.Model):
    def to_dict(self):
        return {column.key: getattr(self, attr) for attr, column in self.__mapper__.c.items()}


db = flask_sqlalchemy.SQLAlchemy(
    model_class=BaseModel,
    session_options={'expire_on_commit': True},
)



import os
import pathlib

import git.exc


def get_tag(prefix=''):
    if prefix and not prefix.endswith('/'):
        prefix = f'{prefix}/'
    try:
        repo = git.Repo('.')
        name = pathlib.Path(repo.remote('origin').url)
        tag = ':'.join(
            [
                name.with_suffix('').name,
                repo.head.commit.hexsha,
            ]
        )
        tag = f'{prefix}{tag}'
        return {
            'tag': tag,
            'commit': repo.head.commit.hexsha,
            'branch': None if repo.head.is_detached else repo.active_branch.name,
        }
    except git.exc.InvalidGitRepositoryError:
        tag = ':'.join(
            [
                os.path.basename(os.getcwd()),
                'latest',
            ]
        )
        return {'tag': f'{prefix}/{tag}'}



"""
Client for interfacing with podman.

To enable using podman as the container interface, specify the following section
in the ``teststack.toml``

.. code-block:: toml

    [client]
    name = "podman"

There is an extra provided to install the ``podman`` dependency.

.. code-block:: shell

    python3 -m pip install teststack[podman]
"""
import json
import os
import platform
import socket
import subprocess

import click
import podman.errors


class Client:
    def __init__(self, machine_name=None, **kwargs):
        if machine_name is not None:
            kws = self._get_connection(machine_name)
            kws.update(kwargs)
        elif platform.system() == 'Darwin':
            kws = self._get_connection('*')
            kws.update(kwargs)
        else:
            kws = kwargs

        if kws:
            self.client = podman.PodmanClient(**kws)
        else:
            self.client = podman.from_env()

    def _process_image_shortname(self, name, default='docker.io'):
        if '/' in name:
            domain = name.split('/')[0]
            try:
                socket.getaddrinfo(domain, 443)
                return name
            except socket.gaierror:
                pass
        return f'docker.io/{name}'

    def _get_connection(self, name):
        connections = subprocess.check_output(
            [
                'podman',
                'system',
                'connection',
                'list',
                '--format=json',
            ]
        )

        if connections:
            connections = json.loads(connections)
        else:
            return {}

        for connection in connections:
            if connection['Name'].endswith(name):
                return {
                    'base_url': f'http+{connection["URI"]}',
                    'identity': connection['Identity'],
                }
        return {}

    def end_container(self, name):
        try:
            container = self.client.containers.get(name)
        except podman.errors.NotFound:
            return
        try:
            container.stop()
            container.wait()
        except podman.errors.APIError:
            pass
        finally:
            container.remove(v=True)

    def container_get(self, name):
        try:
            return self.client.containers.get(name).id
        except podman.errors.NotFound:
            return None

    def container_get_current_image(self, name):
        container = self.container_get(name)
        if container:
            return self.client.containers.get(container).image.id
        return None

    def run(
        self,
        name,
        image,
        ports=None,
        command=None,
        environment=None,
        stream=False,
        user=None,
        volumes=None,
        mount_cwd=False,
    ):
        mounts = volumes or []
        if mount_cwd is True:
            mounts.append(
                {
                    'source': os.getcwd(),
                    'target': self.client.images.get(image).attrs['Config']['WorkingDir'],
                    'type': 'bind',
                }
            )

        if command is True:
            command = ['tail', '-f', '/dev/null']

        if ports:
            for port, hostport in ports.items():
                if not hostport:
                    ports[port] = None

        if not self.image_get(image):
            self.client.images.pull(self._process_image_shortname(image))

        container = self.client.containers.create(
            name=name,
            image=image,
            detach=True,
            stream=stream,
            ports=ports or {},
            environment=environment or {},
            command=command,
            mounts=mounts,
        )

        container.start()
        container.wait(condition="running")

        return container.id

    def start(self, name):
        container = self.container_get(name)
        self.client.containers.get(container).start()

    def image_get(self, tag):
        try:
            return self.client.images.get(self._process_image_shortname(tag)).id
        except podman.errors.ImageNotFound:
            return None

    def run_command(self, container, command):
        container = self.client.containers.get(container)
        click.echo(click.style(f'Run Command: {command}', fg='green'))
        socket = container.exec_run(
            cmd=command,
            tty=True,
            socket=True,
        )

        for line in socket.output:
            click.echo(line, nl=False)

    def build(self, dockerfile, tag, rebuild):
        image, _ = self.client.images.build(path='.', dockerfile=dockerfile, tag=tag, nocache=rebuild, rm=True)
        return image.id

    def get_container_data(self, name, inside=False):
        data = {}
        try:
            container = self.client.containers.get(name)
        except podman.errors.NotFound:
            return None
        data['HOST'] = container.attrs['NetworkSettings']['IPAddress'] if inside else 'localhost'
        for port, port_data in container.attrs['NetworkSettings']['Ports'].items():
            if inside:
                data[f'PORT;{port}'] = port.split('/')[0]
            elif port_data:
                data[f'PORT;{port}'] = port_data[0]['HostPort']
        return data

    def exec(self, container):
        os.execvp('podman', ['podman', 'exec', '-ti', container, 'bash'])



import figenv


class Config(metaclass=figenv.MetaConfig):
    POSTGRES_MAIN_USER = 'testapp'
    POSTGRES_MAIN_PASSWORD = 'secret'
    POSTGRES_MAIN_HOST = 'localhost'
    POSTGRES_MAIN_PORT = 5432
    POSTGRES_MAIN_DBNAME = 'app'

    def SQLALCHEMY_DATABASE_URI(cls):  # lgtm [py/not-named-self]
        return 'postgresql://{user}:{password}@{host}:{port}/{dbname}?sslmode=prefer'.format(
            user=cls.POSTGRES_MAIN_USER,
            password=cls.POSTGRES_MAIN_PASSWORD,
            host=cls.POSTGRES_MAIN_HOST,
            port=cls.POSTGRES_MAIN_PORT,
            dbname=cls.POSTGRES_MAIN_DBNAME,
        )

    CACHE_TYPE = 'RedisCache'
    REDIS_MAIN_HOST = 'localhost'
    REDIS_MAIN_PORT = 6379

    def CACHE_REDIS_URL(cls):  # lgtm [py/not-named-self]
        return 'redis://@{host}:{port}/0'.format(
            host=cls.REDIS_MAIN_HOST,
            port=cls.REDIS_MAIN_PORT,
        )



import pathlib
import tempfile
from unittest import mock

from docker.errors import ImageNotFound
from docker.errors import NotFound

from teststack import cli


def test_render(runner, tag):
    with tempfile.NamedTemporaryFile() as tmpfile:
        result = runner.invoke(cli, ['render', f'--dockerfile={tmpfile.name}'])
        assert result.exit_code == 0
        with open(tmpfile.name, 'r') as fh_:
            assert fh_.readline() == 'FROM docker.io/python:slim\n'
            assert fh_.readline() == 'ENV PYTHON=True\n'
            assert fh_.readline() == 'WORKDIR /srv\n'
            assert fh_.readline() == '\n'
            assert 'docker-metadata' in fh_.readline()
            assert tag['commit'] in fh_.readline()


def test_render_isolated(runner):
    with open('Dockerfile.j2') as fh_, runner.isolated_filesystem() as th_:

        with open('Dockerfile.j2', 'w') as wh_:
            wh_.write(fh_.read())

        result = runner.invoke(cli, [f'--path={th_}', 'render'])
        assert result.exit_code == 0
        with open('Dockerfile', 'r') as fh_:
            assert fh_.readline() == 'FROM docker.io/python:slim\n'
            assert fh_.readline() == 'ENV PYTHON=True\n'
            assert fh_.readline() == 'WORKDIR /srv\n'
            assert not fh_.readline()


def test_container_start_no_tests(runner, attrs):
    client = mock.MagicMock()
    client.containers.get.return_value.attrs = attrs
    with mock.patch('docker.from_env', return_value=client):
        result = runner.invoke(cli, ['start', '-n'])
    assert client.containers.get.call_count == 4
    assert client.containers.run.called is False
    assert result.exit_code == 0


def test_container_start_no_tests_not_started(runner, attrs):
    client = mock.MagicMock()
    client.containers.get.return_value.attrs = attrs
    client.containers.get.side_effect = NotFound('container not found')
    with mock.patch('docker.from_env', return_value=client):
        result = runner.invoke(cli, ['start', '-n'])
    assert client.containers.get.call_count == 2
    assert client.containers.run.call_count == 2
    assert result.exit_code == 0


def test_container_start_with_tests(runner, attrs):
    client = mock.MagicMock()
    client.containers.get.return_value.attrs = attrs
    client.images.get.return_value.id = client.containers.get.return_value.image.id
    with mock.patch('docker.from_env', return_value=client):
        result = runner.invoke(cli, ['start'])
    assert client.containers.get.call_count == 11
    assert client.containers.run.called is False
    assert result.exit_code == 0


def test_container_start_with_tests_old_image(runner, attrs):
    client = mock.MagicMock()
    client.containers.get.return_value.attrs = attrs
    with mock.patch('docker.from_env', return_value=client):
        result = runner.invoke(cli, ['start'])
    assert client.containers.get.call_count == 11
    assert client.containers.run.called is True
    assert client.containers.get.return_value.stop.called is True
    assert client.containers.get.return_value.wait.called is True
    client.containers.get.return_value.remove.assert_called_with(v=True)
    assert result.exit_code == 0


def test_container_start_with_tests_not_started(runner, attrs):
    client = mock.MagicMock()
    client.containers.get.return_value.attrs = attrs
    client.containers.get.side_effect = NotFound('container not found')
    with mock.patch('docker.from_env', return_value=client):
        result = runner.invoke(cli, ['start'])
    assert client.containers.get.call_count == 6
    assert client.containers.run.call_count == 3
    assert result.exit_code == 0


def test_container_stop(runner, attrs):
    client = mock.MagicMock()
    client.containers.get.return_value.attrs = attrs
    with mock.patch('docker.from_env', return_value=client), mock.patch(
        'teststack.containers.docker.Client.end_container'
    ) as end_container:
        result = runner.invoke(cli, ['stop'])
    assert client.containers.get.call_count == 3
    assert end_container.call_count == 3
    assert result.exit_code == 0


def test_container_stop_without_containers(runner, attrs):
    client = mock.MagicMock()
    client.containers.get.return_value.attrs = attrs
    client.containers.get.side_effect = NotFound('container not found')
    with mock.patch('docker.from_env', return_value=client), mock.patch(
        'teststack.containers.docker.Client.end_container'
    ) as end_container:
        result = runner.invoke(cli, ['stop'])
    assert client.containers.get.call_count == 3
    assert end_container.called is False
    assert result.exit_code == 0


def test_container_build(runner, build_output):
    client = mock.MagicMock()
    client.api.build.return_value = build_output
    with mock.patch('docker.from_env', return_value=client):
        result = runner.invoke(cli, ['build', '--tag=blah'])
    client.api.build.assert_called_with(path='.', dockerfile='Dockerfile', tag='blah', nocache=False, rm=True)
    assert result.exit_code == 0


def test_container_start_with_tests_without_image(runner, attrs):
    client = mock.MagicMock()
    client.containers.get.return_value.attrs = attrs
    image = mock.MagicMock()
    client.images.get.side_effect = [ImageNotFound('image not found'), image, image, image]
    with mock.patch('docker.from_env', return_value=client):
        result = runner.invoke(cli, ['start'])
    assert client.containers.get.call_count == 11
    assert client.containers.run.called is True
    assert client.images.get.call_count == 4
    assert result.exit_code == 0


def test_container_run(runner, attrs):
    client = mock.MagicMock()
    client.containers.get.return_value.attrs = attrs
    client.images.get.return_value.id = client.containers.get.return_value.image.id
    client.containers.get.return_value.exec_run.return_value.output = [
        'foo',
        'bar',
        'baz',
    ]
    with mock.patch('docker.from_env', return_value=client):
        result = runner.invoke(cli, ['run'])
    assert client.containers.get.call_count == 14
    assert client.containers.run.called is False
    assert result.exit_code == 0
    assert 'foobarbaz' in result.output
    assert 'Run Command: env' in result.output


def test_container_run_step(runner, attrs):
    client = mock.MagicMock()
    client.containers.get.return_value.attrs = attrs
    client.images.get.return_value.id = client.containers.get.return_value.image.id
    client.containers.get.return_value.exec_run.return_value.output = [
        'foo',
        'bar',
        'baz',
    ]
    with mock.patch('docker.from_env', return_value=client):
        result = runner.invoke(cli, ['run', '--step=install'])
    assert client.containers.get.call_count == 13
    assert client.containers.run.called is False
    assert result.exit_code == 0
    assert 'foobarbaz' in result.output
    assert 'Run Command: env' not in result.output
    assert 'Run Command: python -m pip install' in result.output



import os
import subprocess
from unittest import mock

import pytest
from docker.errors import NotFound

from teststack import cli


def test_running_tests_in_containers(docker, runner, testapp_dir):

    result = runner.invoke(cli, [f'--path={testapp_dir}', 'stop', 'render', 'build', 'run'])
    assert result.exit_code == 0
    result = runner.invoke(cli, [f'--path={testapp_dir}', 'stop'])
    assert result.exit_code == 0


def test_running_tests(docker, runner, testapp_dir):

    result = runner.invoke(cli, [f'--path={testapp_dir}', 'stop', 'render', 'build', 'start', '-n'])
    assert result.exit_code == 0
    with pytest.raises(NotFound):
        docker.containers.get('testapp_tests')

    result = runner.invoke(cli, [f'--path={testapp_dir}', 'env', '-n'])
    assert result.exit_code == 0
    envvars = dict(map(lambda x: x.split('='), result.output.strip().split('\n')))
    envvars['PYTHONPATH'] = str(testapp_dir)
    with mock.patch.dict(os.environ, envvars):
        subprocess.run(['poetry', 'install'], cwd=testapp_dir)
        result = subprocess.run(['poetry', 'run', 'pytest'], cwd=testapp_dir)
        assert result.returncode == 0

    result = runner.invoke(cli, [f'--path={testapp_dir}', 'stop'])
    assert result.exit_code == 0



from flask import current_app
from flask import jsonify
from flask.views import MethodView

from app.database import db


def health_database_status():
    try:
        # to check database we will execute raw query
        db.session.execute('SELECT 1')
    except Exception as e:
        return False
    return True


def health_cache_status():
    for cache, backend in current_app.extensions['cache'].items():
        try:
            backend._read_clients.keys()
        except Exception as e:
            return False
    return True


class Health(MethodView):
    def get(self):
        database = health_database_status()
        cache = health_cache_status()
        return jsonify(
            {
                'status': database and cache,
                'database': database,
                'cache': cache,
            }
        )



from teststack import cli


def test_env_with_containers(runner, testapp_dir):
    result = runner.invoke(cli, [f'--path={testapp_dir}', 'start', 'env', 'stop'])
    assert result.exit_code == 0
    assert 'AWS_ACCESS_KEY_ID' in result.output
    assert 'POSTGRES_MAIN_DBNAME' in result.output



import json
import os
import select
import socket
import sys

import click
import docker.errors

from ..utils import read_from_stdin


class Client:
    def __init__(self, **kwargs):
        context = docker.ContextAPI.get_current_context()
        if context.name == 'default':
            self.client = docker.from_env()
        else:
            self.client = docker.DockerClient(
                base_url=context.Host,
                tls=context.TLSConfig,
            )

    def end_container(self, name):
        try:
            container = self.client.containers.get(name)
        except docker.errors.NotFound:
            return
        container.stop()
        container.wait()
        container.remove(v=True)

    def container_get(self, name):
        try:
            return self.client.containers.get(name).id
        except docker.errors.NotFound:
            return None

    def container_get_current_image(self, name):
        container = self.container_get(name)
        if container:
            return self.client.containers.get(container).image.id
        return None

    def run(
        self,
        name,
        image,
        ports=None,
        command=None,
        environment=None,
        stream=False,
        user=None,
        volumes=None,
        mount_cwd=False,
    ):
        if mount_cwd is True:
            volumes = volumes or {}
            volumes.update(
                {
                    os.getcwd(): {
                        'bind': self.client.images.get(image).attrs['Config']['WorkingDir'],
                        'mode': 'rw',
                    },
                }
            )
        if command is True:  # pragma: no branch
            command = "sh -c 'trap \"trap - TERM; kill -s TERM -- -$$\" TERM; tail -f /dev/null & wait'"

        return self.client.containers.run(
            name=name,
            image=image,
            detach=True,
            stream=stream,
            user=user,
            ports=ports or {},
            command=command,
            environment=environment or {},
            volumes=volumes,
        ).id

    def start(self, name):
        self.client.containers.get(name).start()

    def image_get(self, tag):
        try:
            return self.client.images.get(tag).id
        except docker.errors.ImageNotFound:
            return None

    def run_command(self, container, command):
        container = self.client.containers.get(container)
        click.echo(click.style(f'Run Command: {command}', fg='green'))
        sock = container.exec_run(
            cmd=command,
            tty=True,
            stdin=True,
            socket=True,
        ).output

        with read_from_stdin() as fd:
            if fd is not None:  # pragma: no cover
                BREAK = False
                while not BREAK:
                    reads, _, _ = select.select([sock._sock, fd], [], [], 0.0)
                    for read in reads:
                        if isinstance(read, socket.socket):
                            line = read.recv(4096)
                            if not line:
                                BREAK = True
                            click.echo(line, nl=False)
                        else:
                            sock._sock.send(sys.stdin.read(1).encode('utf-8'))
            else:
                for line in sock:
                    click.echo(line, nl=False)

    def build(self, dockerfile, tag, rebuild):
        for chunk in self.client.api.build(path='.', dockerfile=dockerfile, tag=tag, nocache=rebuild, rm=True):
            for line in chunk.split(b'\r\n'):
                if not line:
                    continue
                data = json.loads(line)
                if 'stream' in data:
                    click.echo(data['stream'], nl=False)

    def get_container_data(self, name, inside=False):
        data = {}
        try:
            container = self.client.containers.get(name)
        except docker.errors.NotFound:
            return None
        self.start(name)
        data['HOST'] = container.attrs['NetworkSettings']['IPAddress'] if inside else 'localhost'
        for port, port_data in container.attrs['NetworkSettings']['Ports'].items():
            if inside:
                data[f'PORT;{port}'] = port.split('/')[0]
            elif port_data:
                data[f'PORT;{port}'] = port_data[0]['HostPort']
        return data

    def exec(self, container):
        os.execvp('docker', ['docker', 'exec', '-ti', container, 'bash'])
